// This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL).
// Any Device - Driver - Linux
// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/driver/kernel_2_4.c $
// $Rev: 23671 $
// $Date: 2013-10-31 16:23:27 -0500 (Thu, 31 Oct 2013) $
//
//	Provide special handling for things specific to the 2.4 kernel.
//

#include <linux/version.h>



#if	(LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)) && \
	(LINUX_VERSION_CODE <  KERNEL_VERSION(2,5,0))

#define	MODULE
#include "main.h"



/* #defines	**************************************************************/

#ifndef MODULE_LICENSE
	#define	MODULE_LICENSE(s)
#endif



/* Tags	**********************************************************************/

EXPORT_NO_SYMBOLS;
MODULE_AUTHOR("General Standards Corporation, www.generalstandards.com");
MODULE_DESCRIPTION(	"GSC " DRV_DEV_DEVICE_NAME_UPPER
					" driver " DRV_GSC_FILE_VERSION_STR);
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE("GSC " DRV_DEV_DEVICE_NAME_UPPER);



/******************************************************************************
*
*	Function:	dma_sgl_user_mem_init
*
*	Purpose:
*
*		Initialize things in preparation for future kernel access of user
*		memory for SGL DMA operations.
*
*	Arguments:
*
*		dma		The DMA structure for which any initialization should be done.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dma_sgl_user_mem_init(dev_dma_t* dma)
{
	alloc_kiovec(1, &dma->sgl.iobuf);
}



/******************************************************************************
*
*	Function:	dma_sgl_user_mem_map
*
*	Purpose:
*
*		Map user memory to kernel space for an SGL DMA transfer.
*
*	Arguments:
*
*		dma		The DMA structure with the references to the user memory
*				whose references are to be freed.
*
*		xfer	The data for the DMA transfer to be performed.
*
*	Returned:
*
*		0		All went well.
*		<0		An error code for the problem seen.
*
******************************************************************************/

int dma_sgl_user_mem_map(dev_dma_t* dma, gsc_drv_dma_t* xfer)
{
	unsigned long	bytes;
	u32				dpr;
	u32				dpr_flags;
	int				inc;
	long			l;
	u32				ladr;
	unsigned long	offset;
	unsigned long	padr;
	struct page*	page;
	unsigned long	pages;
	int				rw;
	u32				siz;
	int				status;
	void*			vi_adrs;

	for (;;)
	{
		if (dma->sgl.list == NULL)
		{
			// We didn't get memory for the SGL DPR list.
			status			= -ENOMEM;
			xfer->status	= GSC_DMA_SGL_BUILD_FAILED;
			break;
		}

		rw		= (xfer->u.xfer.flags & GSC_DMA_XFER_DIR_TO_HOST) ? READ : WRITE;
		status	= map_user_kiobuf(	rw,
									dma->sgl.iobuf,
									xfer->u.xfer.user_adrs,
									xfer->u.xfer.bytes);

		if (status)
		{
			// We were unable to map the user memory to the kiovec.
			xfer->status	= GSC_DMA_SGL_BUILD_FAILED;
			break;
		}

		dma->sgl.user_mapped	= 1;

		if (dma->sgl.iobuf->nr_pages > dma->sgl.size)
		{
			// This should never happen!
			status					= -ENOMEM;
			xfer->status			= GSC_INTERNAL_ERROR;
			break;
		}

		// Compute the lower DPR bits.
		dpr_flags	= DMA_DPR_DESC_LOC_PCI;

		if (xfer->u.xfer.flags & GSC_DMA_XFER_DIR_TO_HOST)
			dpr_flags	|= DMA_DPR_BOARD_TO_HOST;
		else
			dpr_flags	|= DMA_DPR_HOST_TO_BOARD;

		// Make other computations.
		bytes	= dma->sgl.iobuf->length;
		inc		= (dma->flags & GSC_DMA_OPEN_LADRS_INC) ? 1 : 0;
		ladr	= xfer->u.xfer.local_adrs;
		offset	= dma->sgl.iobuf->offset;
		pages	= dma->sgl.iobuf->nr_pages - 1;
		status	= 0;

		for (l = 0; l <= pages ; l++)
		{
			page	= dma->sgl.iobuf->maplist[l];
			vi_adrs	= page_address(page);
			padr	= __pa(vi_adrs) + offset;
			dpr		= (dma->sgl.list[l].dpr_32 & 0xFFFFFFF0)
					| dpr_flags;

			if (padr > (0xFFFFFFFFUL - PAGE_SIZE + 1))
				status	= -EINVAL;	// Must be a 32-bit address for PLX.

			if (l == pages)
			{
				// This is the last page, which could also be the first.
				siz	= bytes;
				dpr	|= DMA_DPR_END_OF_CHAIN;
			}
			else if (l)
			{
				// These are the middle pages.
				siz	= PAGE_SIZE;
			}
			else
			{
				// This is the first page.
				siz		= PAGE_SIZE - offset;
			}

			offset						= 0;
			bytes						-= siz;
			dma->sgl.list[l].padr_32	= padr;
			dma->sgl.list[l].ladr_32	= ladr;
			dma->sgl.list[l].siz_32		= siz;
			dma->sgl.list[l].dpr_32		= dpr;

			if (inc)
				ladr	+= siz;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	dma_sgl_user_mem_reset
*
*	Purpose:
*
*		Reset things to cleanup for any activity after kernel access of user
*		memory for SGL DMA operations.
*
*	Arguments:
*
*		dma		The DMA structure for which any reset steps should be taken.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dma_sgl_user_mem_reset(dev_dma_t* dma)
{
	if ((dma) && (dma->sgl.iobuf))
		free_kiovec(1, &dma->sgl.iobuf);
}



/******************************************************************************
*
*	Function:	dma_sgl_user_mem_unmap
*
*	Purpose:
*
*		Free all references to any user memory made for an SGL DMA transfer.
*
*	Arguments:
*
*		dma		The DMA structure with the references to the user memory
*				whose references are to be freed.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dma_sgl_user_mem_unmap(dev_dma_t* dma)
{
	if ((dma) && (dma->sgl.iobuf))
		unmap_kiobuf(dma->sgl.iobuf);
}



/******************************************************************************
*
*	Function:	irq_isr
*
*	Purpose:
*
*		Service an interrupt.  For safety sake, each detected interrupt is
*		cleared then disabled.
*
*	Arguments:
*
*		irq		The interrupt number.
*
*		dev_id	The private data we've associated with the IRQ.
*
*		regs	Unused.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void irq_isr(int irq, void* dev_id, struct pt_regs* regs)
{
	irq_isr_common(irq, dev_id);
}



/******************************************************************************
*
*	Function:	mmap_nopage
*
*	Purpose:
*
*		Implement the mmap nopage() procedure.
*
*	Arguments:
*
*		vma		The virtual memory area of interest.
*
*		address	The virtual address of the page being accessed.
*
*		unused	This argument is unused.
*
*	Returned:
*
*		NULL	The request failed.
*		else	The request succeeded.
*
******************************************************************************/

struct page* mmap_nopage(
	struct vm_area_struct*	vma,
	unsigned long			address,
	int						unused)
{
	struct page*	page;

	mmap_nopage_common(vma, address, &page);
	return(page);
}


/******************************************************************************
*
*	Function:	proc_get_info
*
*	Purpose:
*
*		Implement the get_info() service for /proc file system support.
*
*	Arguments:
*
*		page	The data produced is put here.
*
*		start	Records pointer to where the data in "page" begins.
*
*		offset	The offset into the file where we're to begin.
*
*		count	The limit to the amount of data we can produce.
*
*	Returned:
*
*		int		The number of characters written.
*
******************************************************************************/

int proc_get_info(
	char*	page,
	char**	start,
	off_t	offset,
	int		count)
{
	int	eof;
	int	i;

	i	= proc_read(page, start, offset, count, &eof, NULL);
	return(i);
}



#endif
